#! /usr/bin/env node
import { SimpleCLI } from "./simpleCLI.mjs"
import ffmpeg from "fluent-ffmpeg"
import { unlink, createWriteStream } from "fs"
import fs from "fs/promises"
import path from "path"
import { fileURLToPath } from "url"
import { dirname } from "path"
import { createRequire } from "module"
import { exec } from "child_process"
import { promisify } from "util"
const execAsync = promisify(exec)

const require = createRequire(import.meta.url)
const __filename = fileURLToPath(import.meta.url)
const __dirname = dirname(__filename)

const downloadYouTube = async (url, seconds = 15) => {
    try {
        const downloadsDir = await ensureDir("downloads")
        const videoId = url.split("v=")[1]?.split("&")[0]
        const output = path.join(downloadsDir, videoId) + ".mp4"
        console.log(`Running:`)
        const command = `yt-dlp -f mp4 "${url}" --download-sections "*00:00:00-00:00:${seconds}" -o "${output}"`
        console.log(command)
        const { stdout, stderr } = await execAsync(command)
        console.log(stdout)
        if (stderr) throw new Error(`Download failed: ${stderr}`)
        return { success: true, message: stdout }
    } catch (error) {
        return { success: false, error: error.message }
    }
}

const ensureDir = async (dirName) => {
    const dir = path.join(process.cwd(), dirName)
    try {
        await fs.access(dir)
    } catch {
        await fs.mkdir(dir)
    }
    return dir
}

class AutoGifCLI extends SimpleCLI {
    config = {
        duration: 5,
        fps: 10,
        width: 480,
    }

    welcomeMessage = "🎥 AutoGif CLI - Create GIFs from videos and YouTube"
    endpoint = "https://www.googleapis.com/youtube/v3/search"

    async checkCommand(cwd, channelIds) {
        const channelId = channelIds[0]
        this.log(`Seeing if ${channelId} is live...`)
        const status = await this.isLive(
            channelId,
            "AIzaSyCGXWOfyGKo9xkaDZt8DmHYC2j2V1lfvJA",
        )
        this.log(JSON.stringify(status, null, 2))

        if (status.isLive) this.log(`Channel is live!`)
        else this.log("Channel is not live")

        const videoUrl = `https://youtube.com/watch?v=${status.videoId}`

        const downloadResult = await downloadYouTube(videoUrl, 15)
        if (!downloadResult.success) {
            throw new Error(`Failed to download: ${downloadResult.error}`)
        }

        this.log("Download complete, converting to GIF...")
        const outputPath = path.join(
            process.cwd(),
            `live-${status.videoId}.gif`,
        )
        await this.createGifFromYouTube(videoUrl, outputPath)
        this.log(`GIF created at: ${outputPath}`)
    }

    async isLive(channelIdOrUrl, apiKey) {
        let channelId = channelIdOrUrl
        let videoId = ""

        // Extract channel ID from URL if needed
        if (channelIdOrUrl.includes("youtube.com")) {
            const urlParams = new URLSearchParams(channelIdOrUrl.split("?")[1])
            videoId = urlParams.get("v")
            if (!videoId) throw new Error("Invalid YouTube URL")

            // Get channel ID from video
            const response = await fetch(
                `https://www.googleapis.com/youtube/v3/videos?part=snippet&id=${videoId}&key=${apiKey}`,
            )
            const data = await response.json()
            if (!response.ok || !data.items?.[0]) {
                throw new Error("Could not fetch video details")
            }
            channelId = data.items[0].snippet.channelId
        }

        const cacheFile = path.join(__dirname, `live-status-${channelId}.json`)

        try {
            const cachedData = await fs.readFile(cacheFile, "utf-8")
            const { data, timestamp } = JSON.parse(cachedData)
            if (Date.now() - timestamp < 60000) return data
        } catch (err) {}

        const params = new URLSearchParams({
            part: "snippet",
            channelId,
            type: "video",
            eventType: "live",
            key: apiKey,
        })

        const response = await fetch(`${this.endpoint}?${params}`)
        const data = await response.json()

        if (!response.ok) {
            throw new Error(
                `API error: ${data.error?.message || "Unknown error"}`,
            )
        }

        let result
        if (data.items && data.items.length > 0) {
            const stream = data.items[0]
            result = {
                isLive: true,
                videoId: stream.id.videoId,
                title: stream.snippet.title,
                thumbnailUrl: stream.snippet.thumbnails.default.url,
                channelTitle: stream.snippet.channelTitle,
                startTime: stream.snippet.publishedAt,
            }
        } else {
            result = { isLive: false, videoId }
        }

        await fs.writeFile(
            cacheFile,
            JSON.stringify({
                timestamp: Date.now(),
                data: result,
            }),
        )
        return result
    }

    async downloadCommand(cwd, youtubeUrls) {
        const url = youtubeUrls[0]
        await downloadYouTube(url)
    }

    async gifCommand(cwd, videos) {
        if (videos?.length > 0) {
            await this.createGifFromVideo(cwd, videos[0])
            return
        }

        const downloadsDir = await ensureDir("downloads")
        const gifsDir = await ensureDir("gifs")

        const downloads = await fs.readdir(downloadsDir)
        const gifs = await fs.readdir(gifsDir)

        for (const file of downloads) {
            const baseName = path.parse(file).name
            const gifExists = gifs.some((gif) => gif.startsWith(baseName))

            if (!gifExists) {
                this.log(`Creating GIF for ${file}...`)
                await this.createGifFromVideo(cwd, file)
            }
        }
    }

    async castCommand(cwd) {
        console.log("Coming soon")
    }

    async createGifFromVideo(
        cwd,
        inputPath,
        outputPath = null,
        {
            fps = this.config.fps,
            width = this.config.width,
            startTime = 0,
            duration = this.config.duration,
        } = {},
    ) {
        const downloadsDir = await ensureDir("downloads")
        inputPath = path.join(downloadsDir, inputPath)
        inputPath = inputPath
        if (!outputPath) {
            const gifsDir = await ensureDir("gifs")
            const inputPathObj = path.parse(inputPath)
            outputPath = path.join(gifsDir, `${inputPathObj.name}.gif`)
        }

        this.log("Converting to GIF...")
        return new Promise((resolve, reject) => {
            let command = ffmpeg(inputPath)
                .fps(fps)
                .size(`${width}x?`)
                .noAudio()
                .format("gif")

            if (startTime > 0) command = command.seek(startTime)
            if (duration) command = command.duration(duration)

            command
                .output(outputPath)
                .on("end", () => {
                    this.log("GIF created successfully!")
                    resolve(outputPath)
                })
                .on("error", reject)
                .run()
        })
    }
}

if (import.meta.url === `file://${process.argv[1]}`) {
    const cli = new AutoGifCLI()
    cli.executeUsersInstructionsFromShell()
}

export { AutoGifCLI }
